/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Dynamic extensions loading.
 *
 * See extensions.h for more information.
 *
 * Note : This module is NOT thread-safe. All processing must be done from one thread only.
 */
#include "waaad-internal.h"
#include <waaad/waaad.h>

#include <dlfcn.h>	/* We may use libtool's <ltdl.h> later for better portability.... */
#include <string.h>
#include <errno.h>

/* Structure to store the list of extensions from the configuration file */
typedef struct _ext_t {
	char * filename;
	char * conffile;
	void * handler;
	void (*waaad_ext_fini)(void);
	struct _ext_t *next;
} ext_t;

/* Variables to manage the list of extensions */
static ext_t	ext_head;
static ext_t   *ext_tail;

/* The API structure that is passed to all extensions */
static waaad_api_t exported_api;

/* Simple security to not forgot this part */
#define MY_WAAAD_API_VER 1
#if MY_WAAAD_API_VER != WAAAD_API_VER
# error "You must update the API init here"
#endif
/* Initialize the API that is passed to extensions */
static void api_init(void)
{
	exported_api.version = WAAAD_API_VER;
	API_INIT_LOG        ( exported_api.log        );
	API_INIT_DICTIONARY ( exported_api.dictionary );
	API_INIT_CONF       ( exported_api.conf       );
	API_INIT_PEER       ( exported_api.peer       );
	API_INIT_MSG        ( exported_api.msg        );
	API_INIT_RT         ( exported_api.rt         );
	API_INIT_SESSION    ( exported_api.session    );
	API_INIT_DISP       ( exported_api.dispatch   );
	API_INIT_SEC        ( exported_api.security   );
}

/* Initialize the extensions module */
int ext_init( void ) 
{
	memset(&ext_head, 0, sizeof(ext_t));
	ext_tail = &ext_head;
	api_init();
	TRACE_DEBUG (INFO, "Extension module initialized");
	return 0;
}
	
/* Add a new extension in the list, for loading later */
int ext_register( char * filename, char * conffile )
{
	TRACE_DEBUG (FULL, "Adding %s", filename);
	/* Check the filename is valid */
	if (filename == NULL) {
		TRACE_DEBUG (INFO, "Empty extension filename");
		return EINVAL;
	}
	/* Check that ext_load has not yet been called */
	if (ext_tail->handler != NULL) {
		TRACE_DEBUG (INFO, "ext_register cannot be called after ext_load");
		return EBUSY;
	}
	/* Create a new object in the list */
	ext_tail->next = malloc( sizeof(ext_t) );
	if ( ext_tail->next == NULL) {
		TRACE_DEBUG (INFO, "malloc failed");
		return ENOMEM;
	}
	memset(ext_tail->next, 0, sizeof(ext_t));
	/* Copy the filename in the object */
	ext_tail->next->filename = strdup(filename);
	if ( ext_tail->next->filename == NULL) {
		TRACE_DEBUG (INFO, "strdup failed");
		free(ext_tail->next);
		ext_tail->next = NULL;
		return ENOMEM;
	}
	if (conffile) {
		ext_tail->next->conffile = strdup(conffile);
		if ( ext_tail->next->conffile == NULL) {
			TRACE_DEBUG (INFO, "strdup failed");
			free(ext_tail->next->filename);
			free(ext_tail->next);
			ext_tail->next = NULL;
			return ENOMEM;
		}
	}
		
	/* we are done */
	ext_tail = ext_tail->next;
	TRACE_DEBUG (INFO, "%s added.", filename);
	return 0;
}

/* Actually load all the registered extensions */
int ext_load(void)
{
	int ret, (*waaad_ext_init)(waaad_api_t *, char *) = NULL;
	
	TRACE_DEBUG (FULL, "Loading extensions...");
	
	/* Loop on all extensions, in FIFO order */
	for (ext_tail = ext_head.next; ext_tail != NULL; ext_tail = ext_tail->next)
	{
		TRACE_DEBUG (INFO, "Loading %s", ext_tail->filename);
		
		/* Load the extension */
		ext_tail->handler = dlopen(ext_tail->filename, RTLD_LAZY | RTLD_LOCAL);
		
		if (ext_tail->handler == NULL) {
			/* An error occured */
			log_error("Loading of extension %s failed:\n %s\n", ext_tail->filename, dlerror());
			TRACE_DEBUG (FULL, "Error occurred, message logged -- dlopen.");
			return EINVAL;
		}
		
		/* Resolve the entry point of the extension */
		waaad_ext_init = ( int (*) (waaad_api_t *, char *) )dlsym( ext_tail->handler, "waaad_ext_init" );
		
		if (waaad_ext_init == NULL) {
			/* An error occured */
			log_error("Invalid extension %s:\n %s\n", ext_tail->filename, dlerror());
			TRACE_DEBUG (FULL, "Error occurred, message logged -- dlsym.");
			return EINVAL;
		}
		
		/* Resolve the exit point of the extension, which is optional for extensions */
		ext_tail->waaad_ext_fini = ( void (*) (void) )dlsym( ext_tail->handler, "waaad_ext_fini" );
		
		if (ext_tail->waaad_ext_fini == NULL) {
			/* Not provided */
			TRACE_DEBUG (FULL, "Extension [%s] has no waaad_ext_fini function.", ext_tail->filename);
		} else {
			/* Provided */
			TRACE_DEBUG (FULL, "Extension [%s] waaad_ext_fini has been resolved successfully.", ext_tail->filename);
		}
		
		/* Now call the entry point to initialize the extension */
		ret = (*waaad_ext_init)( &exported_api, ext_tail->conffile );
		
		if (ret != 0) {
			/* The extension was unable to load cleanly */
			log_error("Extension %s could not be initialized, aborting.\n", ext_tail->filename);
			TRACE_DEBUG (FULL, "Error occurred, message logged -- waaad_ext_init.");
			return ret;
		}
		
		/* Proceed to the next extension */
	}
	
	/* We have finished. */
	return 0;
}


/* Now unload the extensions and free the memory */
int ext_fini( void ) 
{
	int ret=0;
	
	TRACE_DEBUG (FULL, "Unloading extensions...");

	/* Loop on all objects */
	for (ext_tail = ext_head.next; ext_tail != NULL; ext_tail = ext_head.next)
	{
		/* Unlink this element from the list */
		ext_head.next = ext_tail->next;
		
		/* Call the exit point of the extension, if it was resolved */
		if (ext_tail->waaad_ext_fini != NULL) {
			TRACE_DEBUG (FULL, "Calling [%s]->waaad_ext_fini function.", ext_tail->filename);
			(*ext_tail->waaad_ext_fini)();
		}
		
		/* Now unload the extension */
		if (ext_tail->handler) {
			TRACE_DEBUG (FULL, "Unloading %s", ext_tail->filename);
			ret = dlclose (ext_tail->handler);
			if (ret != 0) {
				TRACE_DEBUG (INFO, "Unloading [%s] failed : %s\n", ext_tail->filename, dlerror());
			}
		}
		
		/* Free the object and continue */
		free(ext_tail->filename);
		free(ext_tail->conffile);
		free(ext_tail);
	}
	
	/* We return only the error code of last dlclose call... well, anyway, what do we do if it fails? */
	return ret;
}
			
			

		
		
		
		
		
		
		
			
